@@ -1,7 +1,17 @@
 #ifndef NFTABLES_GMPUTIL_H
 #define NFTABLES_GMPUTIL_H
 
+#include <config.h>
+
+#ifdef HAVE_LIBGMP
 #include <gmp.h>
+#else
+#include <mini-gmp.h>
+/* mini-gmp doesn't come with gmp_printf, so we use our own minimal variant */
+extern int mpz_printf(const char *format, const mpz_t value);
+#define gmp_printf mpz_printf
+#endif
+
 #include <asm/byteorder.h>
 
 enum mpz_word_order {
@@ -35,7 +45,6 @@ extern uint32_t mpz_get_uint32(const mpz_t op);
 extern uint16_t mpz_get_uint16(const mpz_t op);
 extern uint8_t mpz_get_uint8(const mpz_t op);
 
-extern uint64_t mpz_get_be64(const mpz_t op);
 extern uint32_t mpz_get_be32(const mpz_t op);
 extern uint16_t mpz_get_be16(const mpz_t op);
 
