@@ -24,12 +24,36 @@ enum debug_level {
 
 #define INCLUDE_PATHS_MAX	16
 
-extern unsigned int max_errors;
-extern unsigned int numeric_output;
-extern unsigned int ip2name_output;
-extern unsigned int handle_output;
-extern unsigned int debug_level;
-extern const char *include_paths[INCLUDE_PATHS_MAX];
+struct output_ctx {
+	unsigned int numeric;
+	unsigned int stateless;
+	unsigned int ip2name;
+	unsigned int handle;
+	unsigned int echo;
+	FILE *output_fp;
+};
+
+struct nft_cache {
+	bool			initialized;
+	struct list_head	list;
+	uint32_t		seqnum;
+};
+
+struct mnl_socket;
+
+struct nft_ctx {
+	struct mnl_socket	*nf_sock;
+	const char		*include_paths[INCLUDE_PATHS_MAX];
+	unsigned int		num_include_paths;
+	unsigned int		parser_max_errors;
+	unsigned int		debug_mask;
+	struct output_ctx	output;
+	bool			check;
+	struct nft_cache	cache;
+	uint32_t		flags;
+};
+
+#define NFT_CTX_DEFAULT		0
 
 enum nftables_exit_codes {
 	NFT_EXIT_SUCCESS	= 0,
@@ -96,7 +120,7 @@ struct input_descriptor {
 	const char			*name;
 	union {
 		const char		*data;
-		int			fd;
+		FILE			*fp;
 	};
 	unsigned int			lineno;
 	unsigned int			column;
@@ -105,7 +129,29 @@ struct input_descriptor {
 };
 
 struct parser_state;
+struct mnl_socket;
 
-int nft_run(void *scanner, struct parser_state *state, struct list_head *msgs);
+int nft_run(struct nft_ctx *nft, struct mnl_socket *nf_sock,
+	    void *scanner, struct parser_state *state,
+	    struct list_head *msgs);
+
+void ct_label_table_init(void);
+void mark_table_init(void);
+void gmp_init(void);
+void realm_table_rt_init(void);
+void devgroup_table_init(void);
+void realm_table_meta_init(void);
+void xt_init(void);
+
+void ct_label_table_exit(void);
+void mark_table_exit(void);
+void realm_table_meta_exit(void);
+void devgroup_table_exit(void);
+void realm_table_rt_exit(void);
+
+int nft_print(struct output_ctx *octx, const char *fmt, ...)
+	__attribute__((format(printf, 2, 3)));
+int nft_gmp_print(struct output_ctx *octx, const char *fmt, ...)
+	__attribute__((format(printf, 2, 0)));
 
 #endif /* NFTABLES_NFTABLES_H */
