@@ -31,6 +31,8 @@
 #include <iface.h>
 #include <cli.h>
 
+#include <libmnl/libmnl.h>
+
 #define CMDLINE_HISTFILE	".nft.history"
 
 static const struct input_descriptor indesc_cli = {
@@ -39,8 +41,9 @@ static const struct input_descriptor indesc_cli = {
 };
 
 static struct parser_state *state;
+static struct nft_ctx *cli_nft;
+static struct mnl_socket *cli_nf_sock;
 static void *scanner;
-
 static char histfile[PATH_MAX];
 static char *multiline;
 static bool eof;
@@ -57,6 +60,10 @@ static char *cli_append_multiline(char *line)
 	}
 
 	len = strlen(line);
+
+	if (len == 0)
+		return NULL;
+
 	if (line[len - 1] == '\\') {
 		line[len - 1] = '\0';
 		len--;
@@ -127,12 +134,13 @@ static void cli_complete(char *line)
 	xfree(line);
 	line = s;
 
-	parser_init(state, &msgs);
+	parser_init(cli_nf_sock, &cli_nft->cache, state, &msgs,
+		    cli_nft->debug_mask, &cli_nft->output);
 	scanner_push_buffer(scanner, &indesc_cli, line);
-	nft_run(scanner, state, &msgs);
-	erec_print_list(stdout, &msgs);
+	nft_run(cli_nft, cli_nf_sock, scanner, state, &msgs);
+	erec_print_list(&cli_nft->output, &msgs, cli_nft->debug_mask);
 	xfree(line);
-	cache_release();
+	cache_release(&cli_nft->cache);
 	iface_cache_release();
 }
 
@@ -141,36 +149,13 @@ static char **cli_completion(const char *text, int start, int end)
 	return NULL;
 }
 
-void __fmtstring(1, 0) cli_display(const char *fmt, va_list ap)
-{
-	int point, end;
-	char *buf;
-
-	point = rl_point;
-	end   = rl_end;
-	rl_point = rl_end = 0;
-
-	rl_save_prompt();
-	rl_clear_message();
-
-	if (vasprintf(&buf, fmt, ap) < 0)
-		fprintf(rl_outstream, "cli_display: out of memory\n");
-	else {
-		fprintf(rl_outstream, "%s\n", buf);
-		xfree(buf);
-	}
-
-	rl_restore_prompt();
-
-	rl_point = point;
-	rl_end   = end;
-	rl_forced_update_display();
-}
-
-int cli_init(struct parser_state *_state)
+int cli_init(struct nft_ctx *nft, struct mnl_socket *nf_sock,
+	     struct parser_state *_state)
 {
 	const char *home;
 
+	cli_nf_sock = nf_sock;
+	cli_nft = nft;
 	rl_readline_name = "nft";
 	rl_instream  = stdin;
 	rl_outstream = stdout;
